@@ -245,3 +245,19 @@ class InvalidPipelineStack(RedisClusterException):
     """
 
     pass
+
+
+class ExternalAuthProviderError(ConnectionError):
+    """
+    Raised when an external authentication provider returns an error.
+    """
+
+    pass
+
+
+class IncorrectPolicyType(Exception):
+    """
+    Raised when a policy type isn't matching to any known policy types.
+    """
+
+    pass
